/*
* Copyright (C) 2000 Arne Schirmacher <arne@schirmacher.de>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*
 * Initial main.c file generated by Glade. Edit as required.
 * Glade will not overwrite this file.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_EXECINFO_H
#include <execinfo.h>
#endif

#include <gnome.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef ENABLE_NLS
#include <libintl.h>
#endif
#include <glade/glade.h>
#include "commands.h"
#include "message.h"

GtkWidget *main_window;
GladeXML *magick_glade;
GladeXML *kino_glade;

static int sigpipe = 0;

void sigpipe_clear( void );
int sigpipe_get( void );

void sigpipe_clear( void )
{
	sigpipe = 0;
}

int sigpipe_get( void )
{
	return sigpipe;
}

static void sigpipe_handler( int value )
{
	printf( "SIGPIPE Received (%d) - setting indicator\n", value );
	sigpipe ++;
}

static void sigusr2_handler( int value )
{
	printf( "SIGUSR2 Received (%d) - scanning for jog/shuttle controllers\n", value );
	startJogShuttle();
}

/**
 * Segmentation fault handler.
 *
 * <p>This handler unwinds the stack and prints the stackframes.</p>
 *
 */
#ifdef HAVE_EXECINFO_H
static void sigsegv_handler()
{
	void *array[ 10 ];
	size_t size;
	char **strings;
	size_t i;

	printf( "\nKino experienced a segmentation fault.\n"
		"Dumping stack from the offending thread\n\n" );
	size = backtrace( array, 10 );
	strings = backtrace_symbols( array, size );

	printf( "Obtained %zd stack frames.\n", size );

	for ( i = 0; i < size; i++ )
		printf( "%s\n", strings[ i ] );

	free( strings );

	printf( "\nDone dumping - exiting.\n" );
	exit( 1 );
}
#endif



int
main ( int argc, char *argv[] )
{
	gchar * filename;
	GnomeProgram *gnome;

#ifdef ENABLE_NLS
	bindtextdomain ( PACKAGE, DATADIR "/locale" );
	bind_textdomain_codeset( PACKAGE, "UTF-8" );
	textdomain ( PACKAGE );
#endif

	signal( SIGPIPE, sigpipe_handler );
	signal( SIGUSR2, sigusr2_handler );
#ifdef HAVE_EXECINFO_H
	signal( SIGSEGV,  sigsegv_handler );
#endif

	g_thread_init( NULL );
	gdk_threads_init();
	gnome = gnome_program_init( PACKAGE, VERSION, LIBGNOMEUI_MODULE, argc, argv,
	                            GNOME_PARAM_APP_DATADIR, DATADIR,
	                            GNOME_PARAM_NONE );
	gdk_rgb_init();
	glade_init();
	filename = gnome_program_locate_file ( gnome,
	                                       GNOME_FILE_DOMAIN_APP_DATADIR,
	                                       "kino/magick.glade", TRUE, NULL );
	if ( filename == NULL )
	{
		modal_message( _( "Failed to locate Glade XML resources." ) );
		exit( 1 );
	}
	magick_glade = glade_xml_new( filename, NULL, NULL );
	g_free( filename );
	glade_xml_signal_autoconnect( magick_glade );

	filename = gnome_program_locate_file ( gnome,
	                                       GNOME_FILE_DOMAIN_APP_DATADIR,
	                                       "kino/kino.glade", TRUE, NULL );
	if ( filename == NULL )
	{
		modal_message( _( "Failed to locate Glade XML resources." ) );
		exit( 1 );
	}
	kino_glade = glade_xml_new( filename, NULL, NULL );
	g_free( filename );
	glade_xml_signal_autoconnect( kino_glade );
	main_window = glade_xml_get_widget( kino_glade, "main_window" );
	gtk_widget_set_events( main_window, GDK_KEY_RELEASE_MASK );
	gtk_window_set_default_icon_from_file( DATADIR "/kino/kino.png", NULL );
	kinoInitialise( main_window );
	gtk_widget_show_all( main_window );
	kinoPostInit();
	gdk_threads_enter();
	newFile();
	bulkLoad( argc, argv );
	gtk_main ();
	gdk_threads_leave();
	return 0;
}
